Set Implicit Arguments.

Require Export Misc.

(* --------------------------------------------------------------- *)
(** * Constructions and basic definitions for F_sub.               *)
(* --------------------------------------------------------------- *)

(* --------------------------------------------------------------- *)
(** ** Inductive type for type expressions.                        *)
(* --------------------------------------------------------------- *)

Inductive T (X : Set) : Set :=
  | eta : X -> T X
  | top : T X
  | arr : T X -> T X -> T X
  | uni : T X -> T (option X) -> T X.

(* --------------------------------------------------------------- *)
(** Functoriality.                                                 *)
(* --------------------------------------------------------------- *)

Fixpoint fct (X Y : Set) (f : X -> Y)
    (x : T X) {struct x} : T Y :=
  match x with
  | eta u => eta (f u)
  | top => top Y
  | arr s t => arr (fct f s) (fct f t)
  | uni s t => uni (fct f s) (fct (optfct f) t)
  end.

Definition shift (X : Set) : T X -> T (option X) :=
  fct (@Some X).

Definition shift_r (X Y : Set) : T (X + Y) -> T (X + option Y) :=
  fct (@sumfct_r X Y (option Y) (@Some Y)).

(* --------------------------------------------------------------- *)
(** ** Environments.                                               *)
(* --------------------------------------------------------------- *)

(* --------------------------------------------------------------- *)
(** Absolute environments.                                         *)
(* --------------------------------------------------------------- *)

Definition env0 (X : Set) : Set := X -> T X.

(* --------------------------------------------------------------- *)
(** Relative environments.                                         *)
(* --------------------------------------------------------------- *)

Definition env (X Y : Set) : Set := Y -> T (X + Y).

Definition env_of_env0 X (g : env0 X) : env empty X :=
  fun x : X => fct (@to_empty_l _) (g x).

(* --------------------------------------------------------------- *)
(** Empty environment.                                             *)
(* --------------------------------------------------------------- *)

Definition eenv (X : Set) : env X empty := empty_inc _.

(* --------------------------------------------------------------- *)
(** Push an absolute environment along a bijection.                *)
(* --------------------------------------------------------------- *)

Definition env0_push X Y (f : Bij X Y) : env0 X -> env0 Y :=
  fun g y => fct f (g (bij_inv f y)).

(* --------------------------------------------------------------- *)
(** Add a binding to an environment.                               *)
(* --------------------------------------------------------------- *)

Definition consenv (X : Set) (e : env0 X) (t : T X) :
  env0 (option X) :=
  fun u : option X =>
    shift
    (match u with
    | None => t
    | Some x => e x
    end).

(* --------------------------------------------------------------- *)
(** Relative version of consenv.                                   *)
(* --------------------------------------------------------------- *)

Definition cenv X Y : env X Y -> T (X + Y) -> env X (option Y) :=
  fun d p y => fct (sumfct_r (@Some Y)) (default d p y).

(* --------------------------------------------------------------- *)
(** Join a relative environment with an absolute environment.      *)
(* --------------------------------------------------------------- *)

Definition appendenv (X Y : Set) (g : env0 X)
  (d : env X Y) : env0 (X + Y) :=
  fun u : X + Y =>
    match u with
    | inl x => fct (@inl _ _) (g x)
    | inr y => d y
    end.

(* --------------------------------------------------------------- *)
(** ** The [sub] relation                                          *)
(* --------------------------------------------------------------- *)

(* --------------------------------------------------------------- *)
(** The following relation is the subtyping minus the requirement
    of well-formedness of the environment.                         *)
(* --------------------------------------------------------------- *)

Inductive sub : forall X : Set,
     env0 X -> T X -> T X -> Prop :=
  | sa_top : forall X g s, sub g s (top X)
  | sa_refl_tvar : forall (X : Set) (x : X) g,
      sub g (eta x) (eta x)
  | sa_trans_tvar : forall (X : Set) g (x : X) (t : T X),
      sub g (g x) t -> sub g (eta x) t
  | sa_arrow  : forall (X : Set) g (s1 t1 s2 t2 : T X),
      sub g t1 s1 -> sub g s2 t2 -> sub g (arr s1 s2) (arr t1 t2)
  | sa_all : forall (X : Set) g
      (t1 s1 : T X) (s2 t2 : T (option X)),
      sub g t1 s1 -> sub (consenv g t1) s2 t2 ->
        sub g (uni s1 s2) (uni t1 t2).

Hint Constructors sub.
